{
 "cells": [
  {
   "attachments": {},
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "# Singly Linked Lists"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "Singly linked lists are a type of data structure that is made up of nodes that are created using self-referential structures. Each node contains a value and a reference to the next node in the sequence. The first node is called the **head** and the last node is called the **tail**. The tail node points to a null reference. Singly linked lists are used to implement other data structures such as stacks, queues, and graphs.\n",
    "\n",
    "Singly linked lists are useful because they allow for constant time insertion and deletion of elements at the beginning of the list. However, they do not allow for constant time access to elements in the middle of the list. To access an element in the middle of the list, you must traverse the list from the head node to the desired node.\n",
    "\n",
    "In this notebook, we will implement a singly linked list in Python and demonstrate how to perform common operations on the list such as insertion, deletion, and traversal.\n"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "\n",
    "```{figure} https://i.ibb.co/mJxcf4k/Screen-Shot-2024-04-09-at-8-22-55-AM.png\n",
    "---\n",
    "name: singly_linked_list\n",
    "width: 80%\n",
    "---\n",
    "A singly linked list is a data structure that is made up of nodes that are created using self-referential structures. Each node contains a value and a reference to the next node in the sequence.\n",
    "```\n",
    "\n",
    "## Nodes\n",
    "\n",
    "A **`Node`** is a recursive data structure that contains an instance attribute **'`next`'** which refers to another object of type `Node` that is next in the sequence. \n",
    "\n",
    "Instance attributes of `Node` that are NOT '`next`' node_ are often referred to as the _data_ or _cargo_, or _payload_ of the node.\n",
    "\n",
    "For example, a node with data \"Monday\" has a reference to the node with data \"Tuesday\", which has a reference to the node with data \"Wednesday\", and so on.\n",
    "\n",
    "```{figure} https://i.ibb.co/zQpFZ8N/node2.png\n",
    "---\n",
    "name: node\n",
    "width: 30%\n",
    "---\n",
    "A node in a linked list contains a reference to the next node in the sequence.\n",
    "```\n",
    "\n",
    "The code below shows a simple implementation of a Node class:"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 1,
   "metadata": {},
   "outputs": [],
   "source": [
    "class Node:\n",
    "    def __init__(self, value):\n",
    "        self.value = value\n",
    "        self.next = None"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## SinglyLinkedList Class\n",
    "\n",
    "Next, we define a singly linked list class that contains a reference to the head node of the list. The singly linked list class provides methods for inserting, deleting, and traversing the list."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "class SinglyLinkedList:\n",
    "    def __init__(self):\n",
    "        self.head = None"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Traversal\n",
    "\n",
    "To traverse the list, we start at the head node and follow the next references until we reach the end of the list (i.e., the tail node).\n",
    "\n",
    "\n",
    "```python\n",
    "def traverse(self):\n",
    "    \n",
    "    # Start at the head node\n",
    "    current = self.head\n",
    "    \n",
    "    # Traverse the list until the current node is None\n",
    "    while current is not None:\n",
    "        print(current.value)\n",
    "\n",
    "        # Move to the next node\n",
    "        current = current.next\n",
    "```\n",
    "\n",
    "```{figure} https://www.codesdope.com/staticroot/images/ds/link14.gif\n",
    "---\n",
    "name: linkedlist-traversal\n",
    "width: 60%\n",
    "align: center\n",
    "---\n",
    "**Traversal**: Keep setting the current node to the next node until the current node is None.\n",
    "```\n"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Insertion\n",
    "\n",
    "There are three common cases for inserting a new node into a singly linked list:\n",
    "\n",
    "1. **Prepend**: Insert a new node at the beginning of the list.\n",
    "2. **Append**: Insert a new node at the end of the list.\n",
    "3. **Insert**: Insert a new node at a specific index in the list.\n",
    "\n",
    "To insert a new node at the beginning of the list, we create a new node with the desired value and set its next reference to the current head of the list. Then, we update the head of the list to point to the new node.\n",
    "\n",
    "```{figure} https://cdn.devdojo.com/images/june2021/LinkedList_frontInsertion.gif\n",
    "---\n",
    "name: linkedlist-insertfromfront\n",
    "width: 60%\n",
    "align: center\n",
    "---\n",
    "**Prepend**: Inserting a node at the front of a singly linked list.\n",
    "```\n",
    "\n",
    "<img src=\"https://i.ibb.co/F7bkw4k/singly-insert-start.png\"><br/><br/>\n",
    "\n",
    "To insert a new node at the end of the list, we create a new node with the desired value and set its next reference to the current head of the list. Then, we update the head of the list to point to the new node.\n",
    "\n",
    "```{figure} https://cdn.devdojo.com/images/june2021/linkedlist-insertfromback1.gif\n",
    "---\n",
    "name: linkedlist-insertfromback\n",
    "width: 60%\n",
    "align: center\n",
    "---\n",
    "**Append**: Inserting a node at the tail of a singly linked list.\n",
    "```\n",
    "\n",
    "```python \n",
    "def append(self, value):\n",
    "    \n",
    "    # Create a new node with the given value\n",
    "    new_node = Node(value)\n",
    "\n",
    "    \n",
    "    # If the list is empty, set the new node as the head\n",
    "    if self.head is None:\n",
    "        self.head = new_node\n",
    "        return\n",
    "\n",
    "    \n",
    "    # Otherwise, traverse the list to find the last node\n",
    "    last = self.head\n",
    "    while last.next:\n",
    "        last = last.next\n",
    "    \n",
    "\n",
    "    # Set the next reference of the last node to the new node\n",
    "    last.next = new_node\n",
    "```\n",
    "\n",
    "<img src=\"https://i.ibb.co/D9c6JY8/singly-insert-end.png\"><br/><br/>\n",
    "\n",
    "To insert a new node at a specific index in the list, we need to find the node that precedes the index where we want to insert the new node. We then update the next reference of the preceding node to point to the new node, and update the next reference of the new node to point to the node that was previously at the index.\n",
    "\n",
    "```{figure} https://cdn.devdojo.com/images/june2021/LinkedList_IndexInsertion.gif\n",
    "---\n",
    "name: linkedlist-index-insertion\n",
    "width: 60%\n",
    "align: center\n",
    "---\n",
    "**Insert**: Inserting a node at a specific index in a singly linked list.\n",
    "```\n",
    "\n",
    "<img src=\"https://i.ibb.co/BTgQvCy/singly-insert-middle.png\"><br/><br/>\n"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": []
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Deletion\n",
    "\n",
    "Similar to insertion, there are three common cases for deleting a node from a singly linked list:\n",
    "\n",
    "1. **Delete head**: Delete the first node in the list.\n",
    "2. **Delete tail**: Delete the last node in the list.\n",
    "3. **Delete specific value**: Delete a node with a specific value from the list.\n",
    "\n",
    "To delete the head of the list, we simply update the head reference to point to the next node in the list.\n",
    "\n",
    "\n",
    "```{figure} https://cdn.devdojo.com/images/june2021/Delfront.gif\n",
    "---\n",
    "name: linkedlist-deletefromfront\n",
    "width: 60%\n",
    "align: center\n",
    "---\n",
    "**Delete head**: Deleting the first node in a singly linked list.\n",
    "```\n",
    "\n",
    "<img src=\"https://i.ibb.co/LQd2N82/singly-delete-start.png\">\n",
    "<br/><br/>\n",
    "\n",
    "To delete the tail of the list, we need to find the node that precedes the tail node. We then update the next reference of the preceding node to point to null.\n",
    "\n",
    "\n",
    "```{figure} https://cdn.devdojo.com/images/june2021/delBack.gif\n",
    "---\n",
    "name: linkedlist-deletefromback\n",
    "width: 60%\n",
    "align: center\n",
    "---\n",
    "**Delete tail**: Deleting the last node in a singly linked list.\n",
    "```\n",
    "\n",
    "<img src=\"https://i.ibb.co/Qmq9LVb/single-delete-end.png\">\n",
    "<br/><br/>\n",
    "\n",
    "To delete a node with a specific value from the list, we need to find the node that precedes the node to be deleted. We then update the next reference of the preceding node to skip over the node to be deleted.\n",
    "\n",
    "```{figure} https://cdn.devdojo.com/images/june2021/delIndex.gif\n",
    "---\n",
    "name: linkedlist-index-deletion\n",
    "width: 60%\n",
    "align: center\n",
    "---\n",
    "**Delete specific value**: Deleting a node with a specific value from a singly linked list.\n",
    "```\n",
    "\n",
    "To delete a node from the list, we need to find the node that precedes the node to be deleted. We then update the next reference of the preceding node to skip over the node to be deleted.\n",
    "\n",
    "```python\n",
    "def delete(self, value):\n",
    "\n",
    "    # If the list is empty, return\n",
    "    if self.head is None:\n",
    "        return\n",
    "\n",
    "    # If the head node has the value to be deleted, update the head to point to the next node\n",
    "    current = self.head\n",
    "    if current.value == value:\n",
    "        self.head = current.next\n",
    "        \n",
    "    else:\n",
    "        # Traverse the list to find the node with the value to be deleted\n",
    "        while current.next is not None:\n",
    "            if current.next.value == value:\n",
    "\n",
    "                # Update the next reference of the current node to skip over the node to be deleted\n",
    "                current.next = current.next.next\n",
    "                break\n",
    "\n",
    "            current = current.next\n",
    "```\n",
    "\n",
    "<img src=\"https://i.ibb.co/xqWsnGM/singly-delete-middle.png\"><br/><br/>"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "\n",
    "## Analysis\n",
    "\n",
    "The time complexity of inserting or deleting a node from a singly linked list is `O(n)` in the worst case, where n is the number of nodes in the list. This is because we may need to traverse the entire list to find the node to be inserted or deleted.\n",
    "\n",
    "The space complexity of a singly linked list is `O(n)`, where n is the number of nodes in the list. This is because we need to store a reference to each node in the list.\n",
    "\n",
    "<!-- ## Putting It All Together\n",
    "\n",
    "\n",
    "```python\n",
    "class SinglyLinkedList:\n",
    "    \n",
    "    def __init__(self):\n",
    "        self.head = None\n",
    "\n",
    "    def traverse(self):\n",
    "        current = self.head\n",
    "        while current is not None:\n",
    "            print(current.value)\n",
    "            current = current.next\n",
    "    \n",
    "    def insert(self, value):\n",
    "        new_node = Node(value)\n",
    "        new_node.next = self.head\n",
    "        self.head = new_node\n",
    "\n",
    "    def delete(self, value):\n",
    "        current = self.head\n",
    "        if current.value == value:\n",
    "            self.head = current.next\n",
    "        else:\n",
    "            while current.next is not None:\n",
    "                if current.next.value == value:\n",
    "                    current.next = current.next.next\n",
    "                    break\n",
    "                current = current.next\n",
    "\n",
    "```\n",
    "\n",
    "Now that we have defined the node and singly linked list classes, we can put it all together and create a singly linked list with some sample values.\n",
    "\n",
    "```python\n",
    "# Create a singly linked list\n",
    "sll = SinglyLinkedList()\n",
    "\n",
    "# Insert some values into the list\n",
    "sll.insert(1)\n",
    "sll.insert(2)\n",
    "sll.insert(3)\n",
    "\n",
    "# Traverse the list\n",
    "sll.traverse()\n",
    "\n",
    "# Delete a value from the list\n",
    "sll.delete(2)\n",
    "\n",
    "# Traverse the list again\n",
    "sll.traverse()\n",
    "```\n",
    "\n",
    "This code will output the following:\n",
    "\n",
    "```\n",
    "3\n",
    "2\n",
    "1\n",
    "3\n",
    "1\n",
    "```\n",
    "\n",
    "This demonstrates how to create a singly linked list, insert values into the list, delete values from the list, and traverse the list to print out the values. -->"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "# Reverse a linked list\n",
    "\n",
    "def reverse_linked_list(linked_list):\n",
    "    prev = None\n",
    "    current = linked_list.head\n",
    "    while current:\n",
    "        next_node = current.next\n",
    "        current.next = prev\n",
    "        prev = current\n",
    "        current = next_node\n",
    "    linked_list.head = prev\n",
    "    return linked_list\n"
   ]
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "Python 3",
   "language": "python",
   "name": "python3"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.9.6"
  },
  "orig_nbformat": 4
 },
 "nbformat": 4,
 "nbformat_minor": 2
}
